#!/usr/bin/python3
# ========================================================================================================
#    AES67.py (should be called directly from run.py when device is in AES67 mode)
#
#    1 - Read AES67 UCI parameters and generate "daemon.conf" configuration file
#    2 - Start aes67-daemon process initializing it with "daemon-conf" file
#
# ========================================================================================================

import os
import flexa_uci
import json
import subprocess
import requests

class AES67Config:

    def __init__(self):
        self.deamon_proc = None
        self.daemonConfFile="/etc/aes67_daemon.conf"
        self.aes67uciToJson()

    def __del__(self):
        NotImplemented

    def aes67uciToJson(self):
        config = {}
        #audio
        config["playout_delay"] = flexa_uci.uci_get("aes67.audio.playout_delay")
        config["tic_frame_size_at_1fs"] = flexa_uci.uci_get("aes67.audio.tic_frame_size_at_1fs")
        config["max_tic_frame_size"] = flexa_uci.uci_get("aes67.audio.max_tic_frame_size")
        config["sample_rate"] = flexa_uci.uci_get("aes67.audio.sample_rate")
        #network
        config["http_port"] = flexa_uci.uci_get("aes67.network.http_port")
        config["rtsp_port"] = flexa_uci.uci_get("aes67.network.rtsp_port")
        config["http_base_dir"] = flexa_uci.uci_get("aes67.network.http_base_dir")
        config["rtp_mcast_base"] = flexa_uci.uci_get("aes67.network.rtp_mcast_base")
        config["rtp_port"] = flexa_uci.uci_get("aes67.network.rtp_port")
        config["ptp_domain"] = flexa_uci.uci_get("aes67.network.ptp_domain")
        config["ptp_dscp"] = flexa_uci.uci_get("aes67.network.ptp_dscp")
        config["sap_mcast_addr"] = flexa_uci.uci_get("aes67.network.sap_mcast_addr")
        config["sap_interval"] = flexa_uci.uci_get("aes67.network.sap_interval")
        config["mdns_enabled"] = flexa_uci.uci_get("aes67.network.mdns_enabled")
        #status -> sinks and sources
        config["status_file"] = flexa_uci.uci_get("aes67.sinks_sources.status_file")
        #logs
        config["log_severity"] = flexa_uci.uci_get("aes67.logs.log_severity")
        config["syslog_proto"] = flexa_uci.uci_get("aes67.logs.syslog_proto")
        config["syslog_server"] = flexa_uci.uci_get("aes67.logs.syslog_server")
        print(config)
        with open(self.daemonConfFile, 'w') as outfile:
            json.dump(config, outfile)

    def startDaemon(self):
        if os.path.isfile(self.daemonConfFile):
            self.deamon_proc = subprocess.Popen(["/usr/bin/aes67-daemon", "-c", "{}".format(self.daemonConfFile)], shell=False)
        else:
            print("ERROR: file {} is missing".format(self.daemonConfFile))


class AES67DaemonREST:

    def __init__(self):
        self.ipaddress = "127.0.0.1"
        self.port = flexa_uci.uci_get("aes67.network.http_port")

    def __del__(self):
        NotImplemented

    def __getDaemon(self, url):
        r = requests.get('http://{}:{}{}'.format(self.ipaddress,self.port,url))
        if r.status_code == 200:
            return r.json()
        else:
            return r.status_code

    def __setDaemon(self, url, payload):
        r = requests.post('http://{}:{}{}'.format(self.ipaddress,self.port,url), json=payload)
        return r.status_code

    def __putDeamon(self, url, payload):
        r = requests.put('http://{}:{}{}'.format(self.ipaddress,self.port,url), data=json.dumps(payload))
        return r.status_code

    def __deleteDaemon(self, url):
        r = requests.delete('http://{}:{}{}'.format(self.ipaddress,self.port,url))
        return r.status_code

    def getDaemonConfig(self):
        return self.__getDaemon('/api/config')

    def setDaemonConfig(self, payload):
        return self.__setDaemon('/api/config', payload)

    def getPTPConfig(self):
        return self.__getDaemon('/api/ptp/config')

    def setPTPConfig(self, payload):
        return self.__setDaemon('/api/ptp/config', payload)

    def getPTPStatus(self):
        return self.__getDaemon('/api/ptp/status')

    def addRTPSource(self):
        NotImplemented

    def remRTPSource(self):
        NotImplemented

    def getRTPSourceSDPFile(self):
       NotImplemented

    def addRTPSink(self, id, data):
        if 0 <= id <= 63:
            return self.__putDeamon('/api/sink/{}'.format(id),data)
        return "ERROR"

    def remRTPSink(self):
        if 0 <= id <= 63:
            return self.__deleteDaemon('/api/sink/{}'.format(id))
        return "ERROR"

    def getRTPSinkStatus(self, id):
        if 0 <= id <= 63:
            return self.__getDaemon('/api/sink/status/{}'.format(id))
        return "ERROR"

    def getAllRTPSources(self):
        return self.__getDaemon('/api/sources')

    def getAllRTPSinks(self):
        return self.__getDaemon('/api/sinks')

    def getAllStreams(self):
        return self.__getDaemon('/api/streams')

    def getRemoteSources(self, method="all"):
        if method != "mdns" and method != "sap" and method != "all":
            return "ERROR"
        return self.__getDaemon('/api/browse/sources/{}'.format(method))


if __name__=="__main__":
    print("Starting AES67Receiver ...")
    aes67 = AES67DaemonREST()
    print("Get Daemon Config...")
    print(aes67.getDaemonConfig())
    print("Get PTP Config...")
    print(aes67.getPTPConfig())
    print("Get PTP Status...")
    print(aes67.getPTPStatus())
    print("Get RTP Sink 0 Status...")
    print(aes67.getRTPSinkStatus(0))
    print("Get all RTP Sources...")
    print(aes67.getAllRTPSources())
    print("Get all RTP Sinks...")
    print(aes67.getAllRTPSinks())
    print("Get all Streams...")
    print(aes67.getAllStreams())
    print("Get all Remote Sources...")
    print(aes67.getRemoteSources())

    payload = {'rtsp_port': 8854, 'http_base_dir': '/aes67_ui', 'log_severity': 1, 'playout_delay': 1,
               'tic_frame_size_at_1fs': 48, 'max_tic_frame_size': 1024, 'sample_rate': 48000,
               'rtp_mcast_base': '239.1.0.1', 'rtp_port': 5004, 'ptp_domain': 0, 'ptp_dscp': 48,
               'sap_mcast_addr': '239.255.255.255', 'sap_interval': 30, 'syslog_proto': 'none',
               'syslog_server': '255.255.255.254:1234', 'status_file': '/mnt/data/status.json', 'interface_name': 'eth0',
               'mdns_enabled': True, 'mac_addr': '00:08:e1:08:54:61', 'ip_addr': '192.168.1.92'}

    print("Set Deamon Config...")
    print(aes67.setDaemonConfig(payload))
########################################################
#r = requests.get('http://192.168.1.92:8080/api/config')
#print(r.status_code)
#if r.status_code == 200:
#    print(r.json())

#payload = {'rtsp_port': 8854, 'http_base_dir': '/aes67_ui', 'log_severity': 1, 'playout_delay': 1, 'tic_frame_size_at_1fs': 48, 'max_tic_frame_size': 1024, 'sample_rate': 48000, 'rtp_mcast_base': '239.1.0.1', 'rtp_port': 5004, 'ptp_domain': 0, 'ptp_dscp': 48, 'sap_mcast_addr': '239.255.255.255', 'sap_interval': 30, 'syslog_proto': 'none', 'syslog_server': '255.255.255.254:1234', 'status_file': './status.json', 'interface_name': 'eth0', 'mdns_enabled': True, 'mac_addr': '00:08:e1:08:54:61', 'ip_addr': '192.168.1.92'}

#payload = {'domain': 0, 'dscp': 48}
#r = requests.post('http://192.168.1.92:8080/api/config', json=payload)

#print(r.status_code)


#r = requests.get('http://192.168.1.92:8080/api/config')
#print(r.status_code)
#if r.status_code == 200:
#    print(r.json())