import time
import enum
import json
import threading
import os
import logging

logger = logging.getLogger(__name__)

class SourceState(enum.IntEnum):
    STOPPED = 1  # AudioSource is stopped. Not monitoring, not playing...
    MONITORING = 2  # AudioSource is monitoring its source but doesn't play
    PLAYING = 3  # AudioSource is playing its source (and monitoring too)

class ManagerState(enum.IntEnum):
    MAIN_PLAYER   = 0
    BACKUP_PLAYER = 1
    STOP_PLAYERS  = 2

class PlayerSource:
    def __init__(self, player, prio, type, playCallback=None):
        self.player = player
        self.prio = prio
        self.type = type
        self.alive = False
        self.state = SourceState.STOPPED
        self.playCallback = playCallback
        self.connectedSince = 0
        self.playingSince = 0
        self.connectionCount = 0
        self.playCount = 0

    def startPlay(self):
        if self.alive and self.state != SourceState.PLAYING:
            self.player.playAudio()
            self.state = SourceState.PLAYING
            self.playingSince = int(time.time())
            self.playCount += 1
            if self.playCallback is not None:
                self.playCallback(self.player, True)

    def stopPlay(self):
        if self.state != SourceState.STOPPED:
            self.player.stopAudio()
            self.state = SourceState.MONITORING
            if self.playCallback is not None:
                self.playCallback(self.player, False)

    def connect(self):
        self.player.connect()
        self.state = SourceState.MONITORING

    def disconnect(self):
        if self.state != SourceState.STOPPED:
            self.player.disconnect()
        self.state = SourceState.STOPPED

    def shutdown(self):
        self.player.shutdown()

class AudioSourceManager:
    '''
        A class to manage priorites for multiple players. It plays the player with higher priority
        and mantains a second player as backup. The remaining players on the list are disconnected.

        If the two players with higher prioprity are disconnected, the poll() function will try to
        get them connected. The poll() function is responsible to execute the prioritization algorithm.

        The poll() function needs to be called externally and it is also called whenever an event
        occurs on the players.
        '''

    def __init__(self, pollTime=1.0, connectCallback=None, playCallback=None):
        self.pollTime = pollTime
        self.sources = []
        self.mainSource = None
        self.backupSource = None
        self.connectCallback = connectCallback
        self.playCallback = playCallback
        self.ask4reportStatus = True
        self.lock = threading.Lock()
        self.cond = threading.Condition(self.lock)
        self.stopThread = False
        self.workerThread = None
        self.statusThread = None

    def start(self):
        self.workerThread = threading.Thread(target=self._workerThreadFunction)
        self.workerThread.start()
        self.statusThread = threading.Thread(target=self._reportStatus)
        self.statusThread.start()

    def _workerThreadFunction(self):
        logger.info('worker thread starting')

        while not self.stopThread:
            self._refreshSources()

        logger.info('worker thread stopping')

    def _refreshSources(self):

        isAnyPlayerAlive = False
        sourcePlaying = None

        with self.lock:
            self.cond.wait(self.pollTime)

            state = ManagerState.MAIN_PLAYER
            for source in self.sources:

                if source.alive:
                    isAnyPlayerAlive = True

                if source.state == SourceState.PLAYING:
                    sourcePlaying = source

                if state == ManagerState.MAIN_PLAYER:
                    if self.mainSource is None:
                        self.mainSource = source
                        self.mainSource.connect()
                        self.mainSource.startPlay()
                        state = ManagerState.BACKUP_PLAYER
                        logger.debug("Connect and play the main player '{}'".format(self.mainSource.player.id))
                    else:
                        if source.alive:
                            if source != self.mainSource:
                                self.backupSource = self.mainSource
                                self.mainSource = source
                                self.mainSource.startPlay()
                                if self.backupSource is not None:
                                    self.backupSource.stopPlay()
                                logger.debug("Play the main player '{}'".format(self.mainSource.player.id))
                            else:
                                source.startPlay()

                            state = ManagerState.BACKUP_PLAYER
                        else:
                            source.connect()

                elif state == ManagerState.BACKUP_PLAYER:
                    if self.backupSource is None:
                        self.backupSource = source
                        self.backupSource.connect()
                        state = ManagerState.STOP_PLAYERS
                        logger.debug("Connect the backup player '{}'".format(self.backupSource.player.id))
                    else:
                        if source.alive:
                            if source != self.backupSource:
                                curPlayer = self.backupSource
                                self.backupSource = source
                                self.backupSource.connect()
                                logger.debug("Connect the backup player '{}'".format(self.backupSource.player.id))

                            state = ManagerState.STOP_PLAYERS
                        else:
                            source.connect()

                elif state == ManagerState.STOP_PLAYERS:
                    source.disconnect()

            # we need this to update state and counters when there is no more players alive
            if not isAnyPlayerAlive and sourcePlaying is not None:
                sourcePlaying.stopPlay()

    def _playerCallback(self, player, alive):
        logger.info("Received a callback from player '{}': alive {}".format(player.id, alive))
        for source in self.sources:
            if source.player == player:
                logger.info(f"{source.prio}, {source.type}")
                with self.cond:
                    if self.connectCallback is not None and source.alive != alive:
                        self.connectCallback(player, alive)
                    source.alive = alive
                    if alive:
                        source.connectedSince = int(time.time())
                        source.connectionCount += 1
                    self.cond.notify()
                    break

    def _sourcesSortKey(self, source):
        return source.prio

    def addPlayer(self, prio, type, player):
        prio = int(prio)
        for source in self.sources:
            if prio == source.prio:
                raise Exception('Priority value already in use: {}'.format(prio))
            if player == source.player:
                raise Exception('Player already on the list')

        player.setCallback(self._playerCallback)

        with self.cond:
            source = PlayerSource(player, prio, type, playCallback=self.playCallback)
            self.sources.append(source)
            # order by prio
            self.sources.sort(key=self._sourcesSortKey)
            self.cond.notify()

    def shutdown(self):
        self.stopThread = True
        self.ask4reportStatus = False
        print("Audio Source Manager shutdown")
        self.workerThread.join()
        with self.lock:
            print("Audio Source Manager - disconnect all sources")
            for source in self.sources:
                source.stopPlay()
                source.player.shutdown()
                if self.connectCallback is not None and source.alive:
                    self.connectCallback(source.player, False)
                source.alive = False
                self.cond.notify()

    def getSource(self, name):
        for source in self.sources:
            if source.player.id == name:
                return source

        raise IndexError('Audio source not found')

    def _reportStatus(self):
        filesPath = "/tmp/"
        destinationFile = filesPath + "reportStatus.json"
        while self.ask4reportStatus:
            filename = filesPath + "sourcesReport" + str(time.time())
            # dict to write in file
            reportDict = {}

            try:
                # open file to write
                f = open(filename, 'w')

                # list to be written in the file with the status of the sources
                statusList = []
                try:
                    for source in self.sources:
                        sourceDict = source.player.reportStatus()
                        sourceDict["priority"]  = source.prio
                        sourceDict["type"]      = source.type
                        sourceDict["state"]     = "playing" if source.state == SourceState.PLAYING else ("stopped" if source.state == SourceState.STOPPED else "monitoring")
                        sourceDict["media"]     = "available" if source.alive else "not available"
                        sourceDict["counter"]   = source.playCount
                        statusList.append(sourceDict)

                    #print(statusList)

                    reportDict["status"] = statusList
                    reportDict["timestamp"] = time.time()

                    f.write(json.dumps(reportDict))
                    f.close()

                    #change tmp file to destinationFile
                    os.rename(filename, destinationFile)
                    #self.logger.info("Status report file written with success.")

                except Exception as exc:
                    self.logger.error("Error reportingStatus: {}%".format(exc))
            except Exception as exc:
                self.logger.error("Error open file for writing: {}%".format(exc))
            time.sleep(2)
