import logging
import threading
import time
import socket
from threading import *
import ipaddress
import struct

from AudioSource import *
import baco

from BARP.barp import BARP
from BARP.simple_paging_utils import SimplePagingDevice

# install this from following repo.
# repo at https://bitbucket.org/kibix/barp_python/src/master/

from enum import Enum


class PagingState(Enum):
    IDLE = 0
    PAGING = 1


class PagingAudioSource(AudioSource):
    def __init__(self, sourceId, name, volume, parameters, audioDev):
        AudioSource.__init__(self, sourceId, name, volume, parameters, audioDev)

        self.audioDevice        = audioDev
        self.pagingMode         = str(AudioSource.getParameter(self, "paging_mode"))
        self.barpID             = int(AudioSource.getParameter(self, "barp_station_id"))

        if self.pagingMode == "ic paging":
            self.barpAudioPort      = int(AudioSource.getParameter(self, "barp_audio_port"))
            self.barpControlPort    = int(AudioSource.getParameter(self, "barp_control_port"))
            self.barpStatusPort     = int(AudioSource.getParameter(self, "barp_status_port"))
            self.barpIPAddress      = str(AudioSource.getParameter(self, "barp_address"))
            self.streamSampleRate   = str(AudioSource.getParameter(self, "sample_rate"))
            self.streamFormat       = str(AudioSource.getParameter(self, "audio_format"))
            self.streamChannels     = "1"
            self.station_name       = ""
            self.device_type        = SimplePagingDevice.UNKNOWN
            self.urlType            = "udp"
        else: #simple paging
            self.barpAudioPort      = 5555
            self.barpControlPort    = 5556
            self.barpStatusPort     = 5557
            self.barpIPAddress      = "0.0.0.0"
            self.streamSampleRate   = "24000"
            self.streamFormat       = "mulaw"
            self.streamChannels     = "1"
            self.station_name       = "unassigned"  # also referred as group name.
            self.device_type        = SimplePagingDevice.ENDPOINT
            self.urlType            = "rtp"

        self.pagingWaitLimit    = 5
        self.statusInterval     = 5

        self.logger             = logging.getLogger("ipspeaker")
        self.lock               = Lock()
        self.event              = Event()
        self.playerStart        = threading.Condition()
        self.playerStop         = threading.Condition()

        # audio source common parameters
        self.externalCallback   = None
        self.alive              = False

        self.equalizer          = None
        self.equalizerConfig    = AudioSource.getParameter(self, "equalizer")

        #baco player
        self.outputFormat       = baco.FrameFormat(baco.STEREO, baco.I16, 2, 44100)

        self.stopPlayer         = False
        self.stopReading        = False
        self.decoderChain       = None
        self.isMonitPaging      = False
        self.control_thread     = None
        self.status_thread      = None
        self.playerThread       = None
        self.reportSenderID     = None

        # BARP initialization
        self.barpParser = BARP(self.barpID, app_type=self.pagingMode)
        self.paging_state = PagingState.IDLE
        self.start_time = time.time()
        self.run()

    def status_message(self):
        if self.paging_state == PagingState.IDLE:
            state = 0
        elif self.paging_state == PagingState.PAGING:
            state = 2

        status_message_kwargs = {
            "state": state,
            "priority": 0,
            "sender_id": self.barpParser.my_id,
            "volume": 2,
            "config_bgm_channel": 0,
        }
        return self.barpParser.encode(
            message_type=128,
            flags=0,
            sender_id=self.barpParser.my_id,
            receiver_id=0,
            sequence_number=0,
            message_kwargs=status_message_kwargs,
            device_type=self.device_type,
            station_name=self.station_name
        )

    def run(self):
        self.control_thread = Thread(target=self.rcvCommand)
        self.status_thread = Thread(target=self.sendStatusMsg)
        self.playerThread = Thread(target=self.runPlayer)
        self.control_thread.daemon = True
        self.status_thread.daemon = True
        self.playerThread.daemon = True
        self.playerThread.start()
        self.control_thread.start()
        self.status_thread.start()

    def rcvCommand(self):
        """Process to receive control signal from control port.
        Note: Should be a parallel thread
        """
        self.logger.debug("Start receiving control thread")
        control_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        control_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if self.barpIPAddress == '0.0.0.0':
            self.logger.info("Broadcast address")
            control_socket.bind(("", self.barpControlPort))
        else:
            control_socket.bind((self.barpIPAddress, self.barpControlPort))
            if ipaddress.IPv4Address(self.barpIPAddress).is_multicast:
                self.logger.info("Barp MCAST address ... will join MC Group")
                mreq = struct.pack("4sl", socket.inet_aton(self.barpIPAddress), socket.INADDR_ANY)
                control_socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)  # join MC Group

        while not self.event.is_set():
            data, address = control_socket.recvfrom(4096)
            header, message = self.barpParser.decode(data)
            with self.lock:
                if self.paging_state == PagingState.IDLE:
                    if header["message_type"] == 3 and message["message_to_me"]:
                        self.logger.info("Paging: control_mess: {}, sender ID: {}, START PAGING".format(header["message_type"], header["sender_id"]))
                        self.paging_state = PagingState.PAGING
                        self.url = "{}://{}:{}".format(self.urlType, message["multicast_stream_group"], message["port"])
                        with self.playerStart:
                            self.playerStart.notify()
                        self.reportSenderID = header["sender_id"]
                        self.startPagingMonitor()
                elif self.paging_state == PagingState.PAGING:
                    if header["message_type"] == 3 and message["message_to_me"]:
                        self.logger.info("Paging: control_mes {}, sender ID: {} PAGING (continue)".format(header["message_type"], header["sender_id"]))
                        self.start_time = time.time()
                    if header["message_type"] == 8 and message["message_to_me"]:
                        self.logger.info("Paging: control_mes {}, sender ID: {}, STOP PAGING".format(header["message_type"], header["sender_id"]))
                        self.stopReading = True
                        self.isMonitPaging = False
                        self.paging_state = PagingState.IDLE
                        self.reportSenderID = None
                        with self.playerStop:
                            if self.alive:
                                self.playerStop.wait()
                        
        self.logger.debug("Stop receiving control message!")

    def sendStatusMsg(self):
        """Process to send state specific status message to master.
        Note: Should be a parallel thread
        """
        statusIPAddress = self.barpIPAddress
        self.logger.debug("Start sending status thread")
        if self.barpIPAddress == "0.0.0.0":
            statusIPAddress = "<broadcast>"
        # initializing port and socket for status reporting.
        status_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        status_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        while not self.event.is_set():
            with self.lock:
                status_socket.sendto(self.status_message(), (statusIPAddress, self.barpStatusPort))
            time.sleep(self.statusInterval)
        self.logger.debug("Stop sending status message!")

    def connect(self):
        pass

    def disconnect(self):
        self.stopAudio()

    def shutdown(self):
        self.logger.info(f"Shutdown source id={self.id}, name={self.name}")
        with self.lock:
            self.isMonitoring = False
            self.isMonitPaging = False
            self.stopPlayer = True

    def playAudio(self):
        with self.lock:
            self.logger.info(f"Playing source id={self.id}, name={self.name}")
            if self.decoderChain is not None:
                self.decoderChain.mute(False)

    def stopAudio(self):
        with self.lock:
            self.logger.info(f"Muting source id={self.id}, name={self.name}")
            if self.decoderChain is not None:
                self.decoderChain.mute(True)

    def runPlayer(self):

        while not self.stopPlayer:

            with self.playerStart:
                self.playerStart.wait()

            streamInput = baco.StreamInput()
            streamInput.setTimeout(5000)

            self.logger.info("Opening {}, format: {}, sample rate: {}, channels: {}".format(self.url, self.streamFormat, self.streamSampleRate, self.streamChannels))
            if self.pagingMode == "ic paging":
                res = streamInput.open(self.url,self.streamFormat,{"sample_rate": self.streamSampleRate,"channels": self.streamChannels})
            else: # simple paging
                self.url = "{}?format={}&channels={}&rate={}".format(self.url, self.streamFormat,self.streamChannels, self.streamSampleRate)
                sdpConf = baco.genSdp(self.url)
                res = streamInput.openSdp(sdpConf)
            self.stopReading = False

            if res == baco.ReturnCode.OK:
                srcInfo = streamInput.getSourceInfo()
                config = baco.Config(srcInfo.getStream(0), self.outputFormat)
                config.inputFifoDuration = 0
                config.outputFifoDuration = 50
                self.decoderChain = baco.DecoderChain(config)
                self.decoderChain.mute(True)
                self.decoderChain.setGain(1,0,0)
                self.decoderChain.setGainMask(1)
                self.setEqualizer()
                self.alsaSink = baco.AlsaSink(self.audioDevice, self.outputFormat, 0, 0)
                if self.equalizer is not None:
                    self.decoderChain.attachSink(self.equalizer)
                    self.equalizer.setFrameSink(self.alsaSink)
                    if self.isEqualizerConfEnabled():
                        self.equalizer.enable(True)
                else:
                    self.decoderChain.attachSink(self.alsaSink)
                self.decoderChain.startOutput()
                if not self.alive:
                    self.alive = True
                    if self.externalCallback is not None:
                        self.externalCallback(self, self.alive)
                while not self.stopReading:
                    res = streamInput.readPacket(self.decoderChain)
                    if res != baco.ReturnCode.OK:
                        self.logger.info("Read Packet Error")
                        self.stopReading = True
                    if self.paging_state == PagingState.IDLE:
                        self.stopReading = True

                self.logger.info("stop player")

                with self.playerStop:
                    self.playerStop.notify()
                    if self.alive:
                        self.alive = False
                        if self.externalCallback is not None:
                            self.externalCallback(self, self.alive)
        self.logger.info("stop player thread")

    def startPagingMonitor(self):
        if not self.isMonitPaging:
            self.isMonitPaging = True
            self.pagingMonThread = Thread(target=self.pagingMonitor)
            self.pagingMonThread.start()

    def pagingMonitor(self):
        self.start_time = time.time()
        while self.isMonitPaging:
            if ( time.time() - self.start_time ) > self.pagingWaitLimit:
                self.logger.info("Paging timeout (5 sec) -- will return to IDLE mode")
                self.stopReading = False
                self.paging_state = PagingState.IDLE
                self.reportSenderID = None
                self.isMonitPaging = False
        if self.alive:
            self.alive = False
            if self.externalCallback is not None:
                self.externalCallback(self, self.alive)

    def reportStatus(self):
        sourceDict = {}
        sourceDict["id"] = self.id
        if self.reportSenderID is not None:
            sourceDict["sender_id"] = self.reportSenderID
        return sourceDict

    def setCallback(self, callback):
        self.externalCallback = callback

    def updateEqualizerConfig(self, name, param, value):
        for filter in self.equalizerConfig['config']:
            if filter['name'] == name:
                filter[param] = str(value)

    def setEqualizer(self):
        try:
            if "config" in self.equalizerConfig and len(self.equalizerConfig['config']) > 0:
                self.logger.info(f"Set Equalizer configuration: {self.equalizerConfig['config']}")
                self.equalizer = baco.Equalizer(self.outputFormat,self.equalizerConfig['config'])
        except Exception as e:
            self.logger.error(e)

    def isEqualizerConfEnabled(self):
        try:
            self.logger.warning(self.equalizerConfig['enable'])
            if self.equalizer is not None and self.equalizerConfig['enable'] == "true":
                self.logger.info("Enable Equalizer")
                return True
            else:
                self.logger.info("Equalizer is disabled (default config)")
                return False
        except Exception as e:
            self.logger.error(e)
            return False

    def setEqualizerParameter(self, config):
        with self.lock:
            try:
                if "config" in self.equalizerConfig and len(self.equalizerConfig['config']) > 0 and any(d["name"] == config["name"] for d in self.equalizerConfig['config']):
                    if "f" in config:
                        if self.equalizer is not None:
                            self.equalizer.setFrequency(config["name"],config["f"])
                        self.updateEqualizerConfig(config['name'],'f',config['f'])
                        self.logger.info(f"Set Filter Central Frequency: {config}")
                    elif "w" in config:
                        if self.equalizer is not None:
                            self.equalizer.setQFactor(config["name"],config["w"])
                        self.updateEqualizerConfig(config['name'],'w',config['w'])
                        self.logger.info(f"Set Filter Q-factor: {config}")
                    elif "g" in config:
                        if self.equalizer is not None:
                            self.equalizer.setGain(config["name"],config["g"])
                        self.updateEqualizerConfig(config['name'],'g',config['g'])
                        self.logger.info(f"Set Filter gain: {config}")
                    else:
                        self.logger.error(f"Set Equalizer parameter:{config}")
                        return False
                    return True

                return False
            except Exception as e:
                self.logger.error(f"Set Equalizer parameter Exception :{e}")
                return False

    def enableEqualizer(self, enable):
        with self.lock:
            try:
                if self.equalizer is not None:
                    self.equalizer.enable(enable)
                self.equalizerConfig['enable'] = enable
                self.logger.info(f"Equalizer enabled {enable}")
            except Exception as e:
                self.logger.error(e)
