
import os
import logging
import random
import time
from AudioSource import *
from threading import *
import baco

class UsbAudioSource(AudioSource):

    def __init__(self, sourceId, name, volume, parameters, audioDev):
        AudioSource.__init__(self, sourceId, name, volume, parameters, audioDev)

        self.audioDevice = audioDev

        self.alive      = False
        self.externalCallback = None
        self.playing    = False
        self.running    = False
        self.lock       = Lock()

        self.logger = logging.getLogger('ipspeaker')
        
        self.fileDir = "/run/media/sda1/"
        self.audioFolder = str(AudioSource.getParameter(self, "path"))
        if self.audioFolder != "":
            self.fileDir += "{}/".format(self.audioFolder)

        self.audioFilePlaying   = None
        self.player             = None
        self.streamAlive        = None
        self.m3uFile            = None
        self.playlist           = None
        self.index              = 0
        self.maxAudioFiles      = 0

        self.stopThread         = False      
        self.workerThread       = Thread(target=self.workerThreadFunction)
        self.workerThread.daemon = True
        self.workerThread.start()        

    def workerThreadFunction(self):
        self.logger.info("worker thread starting")        
        
        while not self.stopThread:
            
            if os.path.isdir(self.fileDir) and not self.alive:
                m3uList = self.getM3UPlaylist()
                if len(m3uList) > 0:
                    m3uPlaylist =  self.fileDir + m3uList[0]
                    self.logger.info("Found playlist file: {}".format(m3uPlaylist))
                    try:
                        f = open(m3uPlaylist,encoding='utf-8',errors='ignore')
                        self.playlist = []
                        for line in f:
                            if line[0] != '#':
                                self.playlist.append(line.strip())
                        self.maxAudioFiles = len(self.playlist)
                        print("M3U Playlist File List ({}):{}".format(self.maxAudioFiles,self.playlist))
                    except Exception as e:
                        self.logger.error("Exception opening playlist: {e}")
                    finally:
                        f.close()

                if not self.maxAudioFiles:
                    self.playlist = self.getMusicFiles()
                    self.maxAudioFiles = len(self.playlist)
                    print("Audio File List ({}):{}".format(self.maxAudioFiles,self.playlist))
                    print("")
                    if self.maxAudioFiles > 0:
                        self.logger.info("Found {} audio files in usb device".format(self.maxAudioFiles))
                        random.shuffle(self.playlist)
                        self.logger.info("Shuffled audio files")
                        print("Shuffled Audio File List ({}):{}".format(self.maxAudioFiles,self.playlist))
                                            
                if self.maxAudioFiles:
                    print("ALIVE")
                    self.logger.info("USB source is alive!")
                    self.alive = True
                    self.playlistPlayer()
                    if self.externalCallback is not None:
                        self.externalCallback(self, self.alive)
                    

            if self.alive and not os.path.isdir(self.fileDir):
                print("DEAD")
                self.logger.info("USB source is dead!")
                self.alive = False
                if self.externalCallback is not None:
                    self.externalCallback(self, self.alive)

            if self.streamAlive is not None:
                if not self.streamAlive:
                    self.playNext()

            time.sleep(0.5)

    def getM3UPlaylist(self):
        return self.checkFiles(self.fileDir,".m3u")

    def getMusicFiles(self):
        musicFiles = []
        musicFiles += self.checkFiles(self.fileDir,".mp3")
        musicFiles += self.checkFiles(self.fileDir,".aac")
        musicFiles += self.checkFiles(self.fileDir,".m4a")
        musicFiles += self.checkFiles(self.fileDir,".flac")
        musicFiles += self.checkFiles(self.fileDir,".ogg")
        musicFiles += self.checkFiles(self.fileDir,".wav")
        return musicFiles

    def checkFiles(self,fileDir, fileExt):
        fileList = []
        for file in os.listdir(fileDir):
            if file.endswith(fileExt):
                fileList.append(file)
        return fileList

    def playlistPlayer(self):
        self.urlAudio = self.fileDir + self.playlist[self.index]
        print(self.urlAudio)
        self.urlAudio = self.urlAudio.replace('&', '_')  # replace unwanted characters
        self.logger.info("Play {}".format(self.urlAudio))
        self.player = baco.SimplePlayer(self.urlAudio, self.audioDevice, 44100)
        self.player.setCallback(self.checkStateCallback)
        self.setEqualizer()
        self.audioFilePlaying = self.playlist[self.index]
        self.player.connectSource()

    def playNext(self):
        self.index +=1
        if self.index == self.maxAudioFiles:
            self.index = 0
        self.playlistPlayer()
        if self.playing:
            self.player.connectAudio()

    def setCallback(self, callback):
        self.externalCallback = callback

    def checkStateCallback(self, alive):
        if alive:
            print("Playing {}".format(self.audioFilePlaying))    
        else:
            print("Stopped playing file... play file")
        self.streamAlive = alive

    def connect(self):
        pass

    def disconnect(self):
        self.stopAudio()

    def shutdown(self):
        self.logger.info("Shutdown USB audio source")
        with self.lock:
            if self.player is not None and self.player.isConnected():
                self.logger.info(f"Disconnecting source id={self.id}, name={self.name}")
                self.player.disconnect()
            self.playing = False
            self.stopThread = True

    def playAudio(self):
        with self.lock:
            self.logger.info(f"Playing source id={self.id}, name={self.name}")
            if self.player is not None:
                self.player.connectAudio()
            self.playing = True

    def stopAudio(self):
        with self.lock:
            self.logger.info(f"Muting source id={self.id}, name={self.name}")
            if self.player is not None:
                self.player.disconnectAudio()
            self.playing = False

    def reportStatus(self):
        sourceDict = {}
        sourceDict["id"] = self.id
        if self.playing:
            sourceDict["playing"] = self.audioFilePlaying
        return sourceDict

    def setEqualizer(self):
        with self.lock:
            AudioSource.setEqualizer(self,self.player)

    def setEqualizerParameter(self, config):
        with self.lock:
            return AudioSource.setEqualizerParameter(self,self.player,config)

    def enableEqualizer(self, enable):
        with self.lock:
            AudioSource.enableEqualizer(self,self.player,enable)
