import json
import logging
import os
import time
import base64

from barix.web.uci import getValueOfUci, setUciConfigs
from barix.web.utils import readFromJSONFile

from restart_services import createListOfServicesToRestart

class InvalidSourceConfigurationError(Exception):
    pass

log = logging.getLogger('flask-backend')

SOURCES_STATUS_REPORT_FILE = "/tmp/reportStatus.json"
MAX_N_SOURCES = 5

def getIPFormerSources():
    try:
        sources = {}
        config_dict = getValueOfUci('flexa_ip_former')
        """
        {'tpa400': {'level': 'PoE_Optimized'}, 'source1': {'name': 'Informacast', 'type': 'informacast', 'config': '{"url":"","idleaudio":"","idleaudio_format":"ULAW_8KHZ"}', 'volume': '90', 'prio': '1', 'enabled': 'true'}, 'source2': {'type': 'syn-apps', 'enabled': 'true', 'volume': '90', 'prio': '2', 'name': 'Synapps', 'config': '{"server":""}'}, 'source3': {}, 'source4': {}, 'source5': {}}
        """
        for key in config_dict:
            if key.startswith('source'):
                if all(k in config_dict[key] for k in ("name", "type", "config", "volume", "prio", "enabled")):
                    source = {"name": config_dict[key]["name"],
                              "type": config_dict[key]["type"],
                              "config": loadSourceConfigJSON(config_dict[key]["config"], config_dict[key]["type"]),
                              "volume": int(config_dict[key]["volume"]),
                              "prio": int(config_dict[key]["prio"]),
                              "enabled": True if config_dict[key]["enabled"] == 'true' else False}
                else:
                    source = {}
                sources[key] = source
    except Exception as e:
        log.error(e)
        raise e
    else:
        return sources


def getSourcesStatus():
    try:
        if os.path.isfile(SOURCES_STATUS_REPORT_FILE):
            sourcesStatus = readFromJSONFile(SOURCES_STATUS_REPORT_FILE)
            return sourcesStatus
    except Exception as e:
        log.error(e)
        raise e


def getPlayerStatus():
    try:
        currStatus = getSourcesStatus()
        if "timestamp" in currStatus:
            timestamp = currStatus["timestamp"]
            if time.time() - timestamp > 3:
                return "Not playing"
            if "status" in currStatus:
                report = currStatus["status"]
                for sourceReport in report:
                    if sourceReport["state"] == "playing":
                        return sourceReport["priority"]
                return ""
        return "Not playing"
    except Exception as e:
        log.error(e)
        raise e


def loadSourceConfigJSON(dictString, sourceType):
    jsonUci = json.loads(dictString)
    for key in jsonUci:
        if key == "equalizer":
            if "enable" in jsonUci["equalizer"].keys():
                jsonUci["equalizer"]["enable"] = True if jsonUci["equalizer"]["enable"] == 'true' else False
        if sourceType == "sip" and key == "passwd":
            jsonUci["passwd"] = base64.b64decode(jsonUci["passwd"].encode("utf-8")).decode("utf-8")
    return jsonUci


def setEqualizerConfigInUci(sourceID, settings):
    try:
        # get source uci config
        sourceConfig = getValueOfUci("flexa_ip_former", "source" + sourceID, "config")
        log.debug("Source {} config: {}".format(sourceID, sourceConfig))
        sourceConfigJSON = json.loads(sourceConfig)
        sourceEqualizer = sourceConfigJSON["equalizer"]
        if "enable" in settings.keys():
            sourceEqualizer["enable"] = settings["enable"]
        if "config" in settings.keys():
            for filter in settings["config"].keys():
                for currFilter in sourceEqualizer["config"]:
                    if currFilter["name"] == filter.upper():
                        for param in settings["config"][filter].keys():
                            currFilter[param] = str(settings["config"][filter][param])
        jsonUcis = { "flexa_ip_former.source" + sourceID + ".config": sourceConfigJSON }
        log.info("Storing equalizer config {}".format(jsonUcis))
        setUciConfigs(jsonUcis, commit=True, restartServices=False)
    except Exception as e:
        log.error(e)
        raise e

def setIpFormerSources(sourcesJSON, restartServices=True):
    #log.debug(sourcesJSON)
    ucisConfigured = []
    ucisList = {}
    try:
        # delete sources that are not in the received data
        for key in sourcesJSON:
            if len(sourcesJSON[key].keys()) == 0:  # empty source dict for instance "source5":{}
                ucisList['flexa_ip_former.' + key + '.type'] = ''
                ucisList['flexa_ip_former.' + key + '.enabled'] = ''
                ucisList['flexa_ip_former.' + key + '.name'] = ''
                ucisList['flexa_ip_former.' + key + '.volume'] = ''
                ucisList['flexa_ip_former.' + key + '.config'] = ''
                ucisList['flexa_ip_former.' + key + '.prio'] = ''
        for i in range(1, MAX_N_SOURCES + 1):
            if 'source'+str(i) not in sourcesJSON.keys():
                ucisList['flexa_ip_former.source' + str(i) + '.type'] = ''
                ucisList['flexa_ip_former.source' + str(i) + '.enabled'] = ''
                ucisList['flexa_ip_former.source' + str(i) + '.name'] = ''
                ucisList['flexa_ip_former.source' + str(i) + '.volume'] = ''
                ucisList['flexa_ip_former.source' + str(i) + '.config'] = ''
                ucisList['flexa_ip_former.source' + str(i) + '.prio'] = ''

        # create ucis for configured sources
        for key in sourcesJSON:
            if len(sourcesJSON[key].keys()) != 0:  # configured sources
                for param in sourcesJSON[key].keys():
                    if param == "type" or param == "volume" or param == "enabled" or param == "prio":
                        ucisList['flexa_ip_former.' + key + '.' + param] = sourcesJSON[key][param]
                    elif param == "name":
                        if sourcesJSON[key][param] == "":
                            ucisList['flexa_ip_former.' + key + '.name'] = '(not configured)'
                        else:
                            ucisList['flexa_ip_former.' + key + '.name'] = sourcesJSON[key][param]
                    elif param == "config":
                        if "equalizer" not in sourcesJSON[key][param].keys() and sourcesJSON[key]["type"] != "informacast" and sourcesJSON[key]["type"] != "syn-apps" and sourcesJSON[key]["type"] != "sip": # equalizer not in sourcesJSON config keys but source type allows to have: probably a new source
                            sourcesJSON[key][param]["equalizer"] = { "enable": False,
                                                                     "config": [ { "name": "LOW", "f": "80", "t": "q", "w": "1", "g": "0"},
                                                                                 {"name": "LOW-MID", "f": "500", "t": "q", "w": "1", "g": "0"},
                                                                                 {"name": "HI-MID", "f": "2000", "t": "q", "w": "1", "g": "0"},
                                                                                 {"name": "HI", "f": "4000", "t": "q", "w": "1", "g": "0"}]}
                        if sourcesJSON[key]["type"] == "sip" and "passwd" in sourcesJSON[key]["config"]:
                            sourcesJSON[key][param]["passwd"] = base64.b64encode(sourcesJSON[key][param]["passwd"].encode("utf-8")).decode("utf-8")

                        ucisList['flexa_ip_former.' + key + '.' + param] = sourcesJSON[key][param]
                    # elif param == "volume":
                    #     ucisList['flexa_ip_former.' + key + '.volume'] = sourcesJSON[key][param]
                    #
                    # elif param == "enabled":
                    #     if sourcesJSON[key][param]:
                    #         ucisList['flexa_ip_former.' + key + '.enabled'] = "true"
                    #     else:
                    #         ucisList['flexa_ip_former.' + key + '.enabled'] = "false"
                    #
                    # elif param == "prio":
                    #     ucisList['flexa_ip_former.' + key + '.prio'] = sourcesJSON[key][param]
                    #
                    # elif param == "config":
                    #     ucisList['flexa_ip_former.' + key + '.config'] = ''
                    else:
                        log.error("Unrecognized source key {}".format(param))
                        raise InvalidSourceConfigurationError

        ucisConfigured = setUciConfigs(ucisList, commit=True, restartServices=restartServices,
                        selectServicesToRestartCallback = createListOfServicesToRestart)
    except Exception as e:
        log.error(e)
        raise e
    else:
        return ucisConfigured