//
// Created by to on 24/02/21.
//

#ifndef BACO_AUDIOLOOPCORE_H
#define BACO_AUDIOLOOPCORE_H

#include <string>
#include <fstream>
#include <mutex>
#include <memory>
#include <cassert>

#include <baco/ReturnCode.h>
#include <baco/NoCopy.h>
#include <baco/ScheduledExecutor.h>

#include <baco/audio/AlsaSink.h>
#include <baco/audio/SilenceSource.h>
#include <baco/audio/FrameFormat.h>
#include <baco/audio/FrameReplicator.h>

#include "InputBlock.h"


/**
 * @startuml "Behavioural model of an Abstract Player"
 *
 *
 * LOOP_STARTING: * Average is not set on resampler
 * LOOP_STARTING: * Instant compensation applied immediately
 *
 * LOOP_RUNNING: * Average compensation set to average of last period
 * LOOP_RUNNING: * Delta is average instant compensation is period
 * LOOP_RUNNING: * resampler->setAvgComp( AVERAGE + DELTA )

 * LOOP_RECOVERING: * Do not modify average compensation
 * LOOP_RECOVERING: * LAST_AVG average of last period
 * LOOP_RECOVERING: * AVERAGE average or current period

 * [*] --> LOOP_STARTING : @init or stop()
 * LOOP_STARTING --> LOOP_RUNNING : abs(comp) < 25
 *
 * LOOP_RUNNING --> LOOP_RECOVERING : XRUN
 *
 * LOOP_RECOVERING --> LOOP_RUNNING : abs(AVERAGE - LAST_AVG) < 10 or abs(DELTA)<2
 *
 * @enduml
 *
 */

enum class LoopMode
{
    LOOP_STARTING ,
    LOOP_RUNNING ,
    LOOP_RECOVERING ,
    SILENT
};


struct AudioLoopStats
{
    LoopMode mode;
    InputBlockStats input;
    struct {
        int64_t sampleCount;
        int     numXruns;
        int64_t tFirstXrun;
    } output;
};


class AudioLoopCore
{
public:
    BACO_NO_COPY(AudioLoopCore)

    /**
     * @param sink ALSA sink
     * @param outFmt frame output format
     * @param compensationInterval interval used to recalculate compensation
     * @param frame replicator module
     */
    AudioLoopCore( baco::AlsaSinkPtr const& sink , baco::FrameFormat const& outFmt, int compensationInterval, double volume, baco::FrameReplicatorPtr const& replicator );

    ~AudioLoopCore() = default;

    /**
     * Set csv file name
     * @param fileName cvs file name
     */
    void setCsvReport( std::string const& fileName );

    /**
     * Set the device into loop mode.
     * @param loopCfg Configuration of the input block
     * @return
     */
    baco::ReturnCode setInputLoop( InputBlockConfig const& loopCfg);

    /**
     * Set the device into silence mode. The input block is destructed CSV
     * report is halted.
     * @return
     */
    baco::ReturnCode disconnectInputLoop();


    AudioLoopStats getStats();

    void iterate();

    void setVolume(int volume);

private:

    void periodComp();


    Poco::Logger&       logger;
    std::mutex          myMutex{};
    InputBlockConfig    myInputConfig{ };
    baco::FrameFormat   myOutFmt;
    bool                myRequestPending{ false };
    LoopMode            myRequestedMode{ LoopMode::SILENT };
    LoopMode            myCurrentMode{ LoopMode::SILENT };
    std::unique_ptr<InputBlock> myInputBlock{ nullptr };
    std::unique_ptr<baco::SilenceSource> mySilenceSource;
    baco::AlsaSinkPtr   myAlsaSink;
    bool                myCsvEnabled{ false };
    std::ofstream       myCsvOfs{ };
    int64_t             myCsvStartTime{0};
    int                 myCompensationInterval;
    double              myVolume;
    baco::FrameReplicatorPtr myFrameReplicator { nullptr };

    int                 myCompCount{ 0 };
    double              myCompAcc{ 0 };
    double              myAvgComp{ 0 };
    double              myRecoveryAvg{ 0 };
    int64_t             myOutSampleCountOffset{ 0 };
    int                 myLastXrunCnt{ 0 };
    int                 myOutXrun{ 0 };
    unsigned long       myTimeFirstOutXrun{ 0 };
    bool                myClearStats{ false };
    baco::ScheduledExecutor myExecutor;
};


typedef std::shared_ptr<AudioLoopCore> AudioLoopCorePtr;


#endif //BACO_AUDIOLOOPCORE_H
